import os
import shutil
from typing import Any, Callable

from django.conf import settings
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.backends.signals import connection_created


class DjangoReadOnlyError(Exception):
    pass


def rebuild_chart_media_cache() -> None:
    """
    Delete and recreate cache if cache is enabled and cache dir exists
    """
    if settings.INSIGHT_CHARTS_USE_MEDIA_CACHE and os.path.exists(
        settings.INSIGHT_MEDIA_CACHE_ROOT
    ):
        shutil.rmtree(settings.INSIGHT_MEDIA_CACHE_ROOT)
        os.mkdir(settings.INSIGHT_MEDIA_CACHE_ROOT)


def should_block(sql: str) -> bool:
    return not sql.lstrip(" \n(").startswith(
        (
            "EXPLAIN ",
            "PRAGMA ",
            "ROLLBACK TO SAVEPOINT ",
            "RELEASE SAVEPOINT ",
            "SAVEPOINT ",
            "SELECT ",
            "SET ",
        )
    ) and sql not in ("BEGIN", "COMMIT", "ROLLBACK")


def blocker(
    execute: Callable[[str, str, bool, dict[str, Any]], Any],
    sql: str,
    params: str,
    many: bool,
    context: dict[str, Any],
) -> Any:
    if should_block(sql):
        msg = "Write queries are currently disabled. Metrics MUST be read_only"
        raise DjangoReadOnlyError(msg)
    return execute(sql, params, many, context)


def install_hook(connection: BaseDatabaseWrapper, **kwargs: object) -> None:
    if blocker not in connection.execute_wrappers:  # pragma: no branch
        connection.execute_wrappers.insert(0, blocker)


class ReadOnly:
    def __init__(self, read_only_metric):
        self.read_only_metric = read_only_metric

    def __enter__(self):
        connection_created.connect(install_hook)
        return self.read_only_metric

    def __exit__(*args, **kwargs):
        connection_created.disconnect(install_hook)


read_only_mode = ReadOnly



from __future__ import annotations

import base64
import io
import os
from dataclasses import dataclass
from enum import Enum

import matplotlib
import matplotlib.dates as mdates
import matplotlib.font_manager as fm
import matplotlib.pyplot as plt
from asgiref.sync import sync_to_async

from django_insights.models import Bucket
from django_insights.settings import settings

matplotlib.use('Agg')
dir_path = os.path.dirname(os.path.realpath(__file__))
fm.fontManager.addfont(f"{dir_path}/static/insights/fonts/ubuntu.ttf")


plt.rcParams['font.family'] = 'sans-serif'
plt.rcParams['font.sans-serif'] = 'Ubuntu'


class ChartSize(Enum):
    SMALL = 'small'
    MEDIUM = 'medium'
    LARGE = 'large'


@dataclass
class ThemeColor:
    primary: str
    face: str
    tick: str
    grid: str
    edge: str


@dataclass
class Theme:
    dark: ThemeColor
    light: ThemeColor


themes = Theme(
    light=ThemeColor(
        primary=settings.INSIGHTS_CHART_LIGHT_PRIMARY_COLOR,
        face="#FEFEFE",
        tick="#333333",
        grid="#333333",
        edge="#FFFFFF",
    ),
    dark=ThemeColor(
        primary=settings.INSIGHTS_CHART_DARK_PRIMARY_COLOR,
        face="#0d1117",
        tick="#FEFEFE",
        grid="#FEFEFE",
        edge="#0d1117",
    ),
)


def get_figsize(size: ChartSize) -> tuple[int, int]:
    """Size of charts

    Args:
        size (ChartSize): SMALL, MEDIUM, LARGE

    Returns:
        tuple[int, int]: return tuple of figure siza
    """

    if size == ChartSize.SMALL:
        return (8, 4)
    if size == ChartSize.MEDIUM:
        return (16, 8)
    if size == ChartSize.LARGE:
        return (32, 16)


def to_bytes_io(fig) -> bytes:
    """Render Mathplotlib figure to file-like object"""
    flike = io.BytesIO()
    fig.savefig(flike, dpi=settings.INSIGHTS_CHART_DPI)

    return flike.getvalue()


def to_base64img(fig) -> str:
    """Render Mathplotlib figure to Base64 encoded image"""
    return base64.b64encode(to_bytes_io(fig)).decode()


def prepare_plot(bucket, theme, size: ChartSize = ChartSize.SMALL):
    """Default plot options, used for all charts"""

    figsize = get_figsize(size)

    fig, ax = plt.subplots(figsize=figsize)
    fig.set_facecolor(theme.face)
    ax.tick_params(axis='x', colors=theme.tick)
    ax.tick_params(axis='y', colors=theme.tick)
    ax.set_facecolor(theme.face)
    ax.set_title(bucket.title)
    ax.set_ylabel(bucket.ylabel)
    ax.set_xlabel(bucket.xlabel)
    ax.yaxis.label.set_color(theme.tick)
    ax.xaxis.label.set_color(theme.tick)
    ax.grid(
        linestyle="dotted",
        linewidth=0.1,
        color=theme.grid,
        zorder=-10,
        visible=False,
    )

    for spine in ax.spines.values():
        spine.set_edgecolor(theme.edge)

    return fig, ax


def render_barchart(yaxis, xaxis, labels, bucket, theme, size) -> str:
    """Render barchart"""
    theme = getattr(themes, theme)
    fig, ax = prepare_plot(bucket, theme, size)
    ax.bar(labels, xaxis, color=theme.primary)

    return fig


def render_hbarchart(yaxis, xaxis, labels, combined_labels, bucket, theme, size) -> str:
    """Render horizontal barchart"""
    theme = getattr(themes, theme)
    fig, ax = prepare_plot(bucket, theme, size)
    bars = ax.barh(
        combined_labels,
        xaxis,
        color=theme.primary,
        align='center',
        height=0.9,
    )

    for bar in bars:
        width = bar.get_width()
        label_y_pos = bar.get_y() + bar.get_height() / 2
        ax.text(width, label_y_pos, s=f'{width}', va='center', size=4)

    for tick in ax.yaxis.get_major_ticks():
        tick.label1.set_fontsize(6)
        tick.label2.set_fontsize(6)

    ax.set_ylabel(bucket.xlabel)
    ax.set_xlabel(bucket.ylabel)

    _, xmax = plt.xlim()
    plt.subplots_adjust(left=0.3, right=0.9)
    plt.xlim(0, xmax + 100)
    return fig


def render_scatterplot(xaxis, yaxis, bucket, theme, size) -> str:
    """Render scatterplot"""
    theme = getattr(themes, theme)

    fig, ax = prepare_plot(bucket, theme, size)
    ax.scatter(xaxis, yaxis, color=theme.primary)

    return fig


def render_timeseries(xaxis, yaxis, bucket, theme, size) -> str:
    """Render timeseries"""
    theme = getattr(themes, theme)

    fig, ax = prepare_plot(bucket, theme, size)
    ax.plot(xaxis, yaxis, '--o', markersize=5, color=theme.primary)

    # Date formatting
    ax.fmt_xdata = mdates.DateFormatter(bucket.xformat)

    return fig


@sync_to_async
def barchart(bucket: Bucket, theme: str, size: str = ChartSize.SMALL) -> str:
    """Barchart"""
    values = bucket.values.all()
    yvalues = [bucket_value.yvalue for bucket_value in values]
    xvalues = [bucket_value.xvalue for bucket_value in values]
    labels = [bucket_value.category for bucket_value in values]

    return render_barchart(yvalues, xvalues, labels, bucket, theme, size)


@sync_to_async
def hbarchart(bucket: Bucket, theme: str, size: str = ChartSize.MEDIUM) -> str:
    """Horizontal Barchart"""
    values = bucket.values.all()
    yvalues = [bucket_value.yvalue for bucket_value in values]
    xvalues = [bucket_value.xvalue for bucket_value in values]
    labels = [bucket_value.category for bucket_value in values]

    combined_labels = [
        f"{bucket_value.category[0:30]}..:{int(bucket_value.yvalue)}"
        for bucket_value in values
    ]

    return render_hbarchart(
        yvalues, xvalues, labels, combined_labels, bucket, theme, size
    )


@sync_to_async
def scatterplot(bucket: Bucket, theme: str, size: str = ChartSize.SMALL) -> str:
    """Scatterplot chart"""
    values = bucket.values.all()
    yvalues = [bucket_value.yvalue for bucket_value in values]
    xvalues = [bucket_value.xvalue for bucket_value in values]

    return render_scatterplot(yvalues, xvalues, bucket, theme, size)


@sync_to_async
def timeseries(bucket: Bucket, theme: str, size: str = ChartSize.SMALL) -> str:
    """Timeseries chart"""
    values = bucket.values.all()

    yvalues = [bucket_value.timestamp for bucket_value in values]
    xvalues = [bucket_value.xvalue for bucket_value in values]

    return render_timeseries(yvalues, xvalues, bucket, theme, size)



from __future__ import annotations

import asyncio
import os
from typing import Any, Callable

from asgiref.sync import sync_to_async
from django.contrib.auth.mixins import LoginRequiredMixin, UserPassesTestMixin
from django.db.models.query import QuerySet
from django.http import HttpResponse
from django.utils.decorators import classonlymethod
from django.views.generic import DetailView, ListView, View
from django_weasyprint import WeasyTemplateResponseMixin

from django_insights.charts import (
    barchart,
    hbarchart,
    scatterplot,
    timeseries,
    to_bytes_io,
)
from django_insights.models import App, Bucket, Counter, Gauge
from django_insights.settings import settings


class CustomPermissionMixin(UserPassesTestMixin):

    """
    Run following permisssion

    class IsSomeAdminUser:
        def has_permission(self, request):
            return
    """

    def run_permissions(self, permissions: list[Callable]) -> list[bool]:
        """
        Run all permission checks from INSIGHT_DASHBOARD_PERMISSIONS settings
        and return output in array
        """
        return [
            permission().has_permission(request=self.request)
            for permission in permissions
            if hasattr(permission(), 'has_permission')
        ]

    def test_func(self):
        return (
            all(self.run_permissions(settings.INSIGHT_DASHBOARD_PERMISSIONS))
            if settings.INSIGHT_DASHBOARD_PERMISSIONS
            else True
        )


class InsightAppMixin(
    LoginRequiredMixin,
    CustomPermissionMixin,
    View,
):
    apps: list[App] = []

    def get_apps(self) -> QuerySet[App]:
        if not self.apps:
            self.apps = App.objects.all()
        return self.apps

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context.update(
            {'app_name': settings.INSIGHTS_APP_NAME, 'apps': self.get_apps()}
        )
        return context


class InsightsAppView(InsightAppMixin, DetailView):
    model = App
    slug_field = 'uuid'
    slug_url_kwarg = 'app_uuid'
    template_name = "insights/app.html"

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        app = self.get_object()

        context = super().get_context_data(**kwargs)
        context.update({'app_label': app.name})
        return context


class InsightsDashboardView(InsightAppMixin, ListView):
    template_name = "insights/dashboard.html"

    def get_queryset(self) -> list[Any]:
        # Get all metrics in one query?
        return Counter.objects.all()

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context.update({'gauges': Gauge.objects.all(), 'app_label': 'Dashboard'})
        return context


@sync_to_async
def get_bucket(bucket_id: int) -> Bucket:
    """Get bucket from async context"""
    return Bucket.objects.get(pk=bucket_id)


class InsightsChartView(LoginRequiredMixin, View):
    @classonlymethod
    def as_view(cls, **initkwargs):
        view = super().as_view(**initkwargs)
        view._is_coroutine = asyncio.coroutines._is_coroutine
        return view

    async def get(self, request, bucket_id):
        theme = self.request.COOKIES.get('theme') or settings.INSIGHTS_DEFAULT_THEME
        bucket: Bucket = await get_bucket(bucket_id)

        filename: str = (
            f"{settings.INSIGHT_MEDIA_CACHE_ROOT}/{bucket.type}-{bucket.pk}-{theme}.png"
        )

        if os.path.exists(filename) and settings.INSIGHT_CHARTS_USE_MEDIA_CACHE:
            with open(filename, 'rb') as cached_image:
                buffer = cached_image.read()

            return HttpResponse(buffer, content_type='image/png')

        fig = None

        if bucket.is_timeseries:
            fig = await timeseries(bucket, theme=theme)
        if bucket.is_scatterplot:
            fig = await scatterplot(bucket, theme=theme)
        if bucket.is_barchart:
            fig = await barchart(bucket, theme=theme)
        if bucket.is_hbarchart:
            fig = await hbarchart(bucket, theme=theme)

        buffer = to_bytes_io(fig)

        if settings.INSIGHT_CHARTS_USE_MEDIA_CACHE:
            with open(filename, 'wb') as cached_image:
                cached_image.write(buffer)

        return HttpResponse(buffer, content_type='image/png')


class InsightsPDFView(WeasyTemplateResponseMixin, InsightAppMixin, ListView):
    """Render PDF"""

    queryset = App.objects.all()

    pdf_stylesheets = [
        settings.STATIC_ROOT + 'insights/css/pdf.css',
    ]

    template_name = 'insights/pdf.html'
    pdf_attachment = False
    pdf_filename = 'foo.pdf'


class InsightsPDFTestView(InsightAppMixin, ListView):
    """Render PDF HTML as test"""

    queryset = App.objects.all()

    pdf_stylesheets = [
        settings.STATIC_ROOT + 'insights/css/pdf.css',
    ]

    def get_template_names(self):
        return ['insights/pdf.html']

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        data = super().get_context_data(**kwargs)

        data.update(
            {
                'test_mode': True,
                'test_stylesheet': 'insights/css/pdf.css',
            }
        )

        return data



from django import template

register = template.Library()


@register.filter
def clean_str(string, char):
    return string.replace(char, '')


@register.filter
def nice_float(string):
    if isinstance(string, float):
        return f'{round(string):,}'.replace(',', '.')
    return string



from asgiref.sync import async_to_sync
from django import template
from django.urls import reverse
from django.utils.safestring import mark_safe

from django_insights.charts import (
    barchart,
    hbarchart,
    scatterplot,
    timeseries,
    to_base64img,
)
from django_insights.models import Bucket
from django_insights.settings import settings

register = template.Library()


@register.simple_tag
def chart(bucket: Bucket, overlay: str = None):
    img_url = reverse("insights:insight_chart", args=[bucket.pk])

    overlay_attr = ""
    if overlay:
        overlay_attr = f'data-hs-overlay="#hs-full-screen-modal-{overlay}"'

    return mark_safe(
        f"""
        <img
            {overlay_attr}
            class="w-full h-auto border-gray-200 cursor-pointer border-t-1"
            src="{img_url}"
            alt="{bucket.question}"
        />
    """
    )


@async_to_sync
async def base64img(bucket: Bucket, override_theme=None):
    theme = override_theme or settings.INSIGHTS_DEFAULT_THEME
    if bucket.is_timeseries:
        fig = await timeseries(bucket, theme=theme)
    if bucket.is_scatterplot:
        fig = await scatterplot(bucket, theme=theme)
    if bucket.is_barchart:
        fig = await barchart(bucket, theme=theme)
    if bucket.is_hbarchart:
        fig = await hbarchart(bucket, theme=theme)

    return to_base64img(fig)


@register.simple_tag
def pdf_chart(bucket_id: int):
    bucket = Bucket.objects.get(pk=bucket_id)
    img_data = base64img(bucket, override_theme="light")

    return mark_safe(
        f"""
        <img
            class="chart"
            src="data:image/png;base64,{img_data}"
        />
    """
    )



import inspect
from collections import namedtuple
from importlib import import_module
from typing import Callable
from unittest.mock import patch

import pytest

App = namedtuple('App', ('module'))


def register_insights_modules(*args, **kwargs) -> dict[str, Callable]:
    """
    Helper function to load insights functions into registry
    This way pytest can look them up and call them from a fixture,
    it also disables collecting parts with mocks

    Returns:
        dict[str, Callable]: registry with functions
    """
    from django.apps import apps

    registry: dict[str, Callable] = {}

    for app_config in apps.get_app_configs():
        for module_to_search in args:
            # Attempt to import the app's module.
            try:
                with patch('django_insights.metrics.InsightMetrics.get_app') as get_app:
                    get_app.return_value = ("foo", App(module="bar"))
                    mod = import_module("%s.%s" % (app_config.name, module_to_search))
                    functions = inspect.getmembers(mod, inspect.isfunction)
                    registry = {**registry, **dict(functions)}
            except Exception:
                continue

    return registry


registry = register_insights_modules("insights")


@pytest.fixture
def insights_query(request) -> Callable | None:
    """
    Returns insights_query fixture, this function
    resolves wrapped function from registry

    Args:
        request (_type_): Current calling context

    Returns:
        Callable | None: Return function to call query
    """
    if func := registry.get(request.node.name.replace("test_", "")):
        return func

    return None


def pytest_runtest_setup(item: pytest.Item) -> None:
    """
    Add insight fixture to insights marked tests

    Args:
        item (pytest.Item): pytest function description
    """
    marker = item.get_closest_marker("insights")
    if marker is None:
        return
    item.fixturenames.insert(0, 'insights_query')
    item.fixturenames.insert(0, 'insights_config')
